#!/usr/bin/env python

import sys
import os
import time
from optparse import OptionParser
import signal
import logging

try:
    exprom_infra_path=os.environ["EXPROM_INFRA_PATH"]
    sys.path.append(exprom_infra_path+"/src/python")
    import exprom_logging
    import exprom_process
except Exception, e:
    logging.warning("Can not locate Mellanox ExpansionROM Infra libraries")
    logging.warning("Some features will be disabled(checkpatch, etc. )")


COVERITY_PLACE = "/.autodirect/app/Coverity/cov-analysis-linux64-7.0.1/bin/"
COVERITY_CONFIGURE = COVERITY_PLACE + "/cov-configure --compiler /usr/bin/gcc --comptype gcc"
COVERITY_MAKE = COVERITY_PLACE +  "/cov-build --dir bin/coverity "
COVERITY_ANALYZE = COVERITY_PLACE + "/cov-analyze --allow-unmerged-emits --dir "
COVERITY_SUBMIT = COVERITY_PLACE + "/cov-commit-defects --host 10.7.0.184 --port 8080 --stream ipxe --user admin --password 3tango -dir "


VERBOSE_MODE=" >/dev/null "
MERGEROM = "./util/mergerom.pl"
ROM_PREFIX_FILE = "arch/i386/prefix/romprefix.S"
OLD_PREFIX_FILE = "arch/i386/prefix/romprefix.S.old"
TEMP_OUTPUT_FILE = "arch/i386/prefix/romprefix.S.new"
GENERAL_FILE = "config/general.h"
LATEST_PXE_DIR = "/mswg/release/exprom/pxe/FlexBoot/latest/"

DEVICES = {
        "4099": "ConnectX3",
        "4103": "ConnectX3-Pro",
        "4113": "ConnectIB"
        }

def openFile(fileName):
    try:
        sourceFile = open(fileName,'r')
    except:
        print "ERROR: Can't open file: %s" % fileName
        cleanup()
        exit(1)
    lines = sourceFile.readlines()
    sourceFile.close()
    return lines

def writeOutput(chunk, fileName):
    try:
        outputFile = open(fileName, 'w')
    except:
        print "ERROR: Can't open file: %s" % BIST_CONF_FILENAME
        cleanup()
        exit(1)
    outputFile.write(chunk)
    outputFile.close()
    return

def configureParser(optParser):
    optParser.add_option("-d", "--device", default="ALL",  help="Device ID. e.g. 4099. (default=ALL)")
    optParser.add_option("-v", "--version", default="3.4.400", help="FlexBoot version")
    optParser.add_option("--release", action="store_true", default=False, help="Build release" )
    optParser.add_option("--test", action="store_true", default=False, help="Add TEST string in the output ROM file name" )
    optParser.add_option("--countlines", action="store_true", default=False, help="Count lines of code and exit" )
    optParser.add_option("-t", "--tag", help="The source GIT tag to build - only when using --git (default is the latest/current)")
    optParser.add_option("-g", "--git", help="The source GIT repository")
    optParser.add_option("--coverity", action="store_true", default=False, help="Build with Coverity code coverage footprint" )
    optParser.add_option("--debug", default="", help="Add debug flags (<file1>:<verbosity>,<file2>:<verbosity>,...)")
    optParser.add_option("--no_checkpatch", action="store_true",default=False, help="Skip checkpatch test before build")
    optParser.add_option("--rom_suffix", action="store_true",default=False, help="Build rom instead of mrom")
    optParser.add_option("--verbose", action="store_true",default=False, help="Verbose output during build")


def checkOptions(options):
    if options.release and options.debug != "":
        print "Can't build release version with debug flags"
        return 1

    if options.device:
        if (not options.device == "ALL") and (not options.device in DEVICES):
            print "Device not found. Please enter a valid device ID or \"ALL\""
            print "Valid devices:"
            for dev in DEVICES.keys():
                print "ID: " + dev + " - Device: " + DEVICES[dev]
            return 1
    else:
        return 1

    if len(options.version.split(".")) != 3:
        print "Invalid version format. Must be A.B.XYZ (e.g. 3.4.160)"
        return 1

    return 0

def updateRomPrefix(deviceId, version):
    os.system("\cp -f " + ROM_PREFIX_FILE + "_ " + ROM_PREFIX_FILE)
    os.system("rm -f " + OLD_PREFIX_FILE)
    os.system("rm -f " + TEMP_OUTPUT_FILE)

    versionList = version.split(".")
    for subVersion in versionList:
        if len ( subVersion ) > 4:
            print "Error romversion %s" % version
            cleanup()
            exit(1)

    prefix = "driver_version:\n .align 16\n .long 0x73786c6d\n .long 0x3a6e6769\n .long 0x0010"

    # Preparing rest of ROM extension
    prefix = prefix + "%04x" % int ( versionList[0] )
    prefix = prefix + "\n .long 0x" +  "%04x" % int ( versionList[1] )
    prefix = prefix + "%04x" % int ( versionList[2] )
    prefix = prefix + "\n .long " + str ( hex ( int ( deviceId ) ) )
    # Adding port number
    prefix = prefix + "00"
    # Adding protocol type
    prefix = prefix + "ff"
    prefix = prefix + "\n\n"

    #open file
    fileLines = openFile(ROM_PREFIX_FILE)

    #create new file
    chunk = ""
    foundRomVersion = False
    for line in fileLines:
        if foundRomVersion == True:
            if line.find("/* ROM version added */") != -1:
                foundRomVersion = False
            else:
                continue
        if line.find("driver_version:") != -1:
            foundRomVersion = True
            chunk = "".join([chunk, prefix])
            continue

        chunk = "".join([chunk, line])

    #write file into temporary file
    writeOutput(chunk, TEMP_OUTPUT_FILE)

    #replace original file
    print "Replacing ROM file"
    rc = os.system("mv -f " + ROM_PREFIX_FILE + " " + OLD_PREFIX_FILE)
    if rc != 0:
        print "Error replacing file 1"
        cleanup()
        exit(rc)

    rc = os.system("mv -f " + TEMP_OUTPUT_FILE + " " + ROM_PREFIX_FILE)
    if rc != 0:
        print "Error replacing file 2"
        cleanup()
        exit(rc)

    return

def build(version, requestedDevices, outputDir, coverityEnabled, debugString, testEnabled):
    for dev in requestedDevices:
	os.system("make clean");
        isFC = (dev == "4099" or dev == "4103")
        print "Current build configuration: Device - " + dev + " ; FC - " + str(isFC)
        updateRomPrefix(dev, version)

        flags = ""
        if isFC:
            flags = "EXTRA_CFLAGS='-DFLASH_CONFIGURATION'"
        # Add debug flag
        debugFlags = ""
        if debugString != "":
            debugFlags = "DEBUG='" + debugString + "'"

        cmd = "make -j `getconf _NPROCESSORS_ONLN`  bin/" + requestedDevices[dev] + ROM_SUFFIX+" " + flags + " " + debugFlags + VERBOSE_MODE

        rc = 0
        if coverityEnabled:
            print cmd
            rc = os.system(COVERITY_MAKE + cmd)
        else:
            print cmd
            rc = os.system(cmd)
        
        if rc != 0:
            print "Build failed!"
            return rc

        # Copy output file to build directory
        fileName = "FlexBoot-" + version + "_"
        if testEnabled:
            fileName += "TEST_"
        if debugFlags != "":
            fileName += "DEBUG_"
        fileName += dev + ROM_SUFFIX

        os.system("\cp -f bin/" + requestedDevices[dev] + ROM_SUFFIX+" " + " " + outputDir + "/" + fileName)

        if coverityEnabled:
            rc = os.system(COVERITY_ANALYZE + "bin/coverity")
            if rc != 0:
                print "Coverity analyze error! Exiting..."
                return rc
            rc = os.system(COVERITY_SUBMIT + "bin/coverity")
            if rc != 0:
                print "Coverity submit error! Exiting..."
                return rc

    return 0

######################################################################
# MAIN
######################################################################
def cleanup():
    # Move back the original files
    os.system("mv -f " + GENERAL_FILE + "_ " + GENERAL_FILE+" 2>/dev/null")
    os.system("mv -f " + ROM_PREFIX_FILE + "_ " + ROM_PREFIX_FILE+" 2>/dev/null")
    os.system("rm -f " + OLD_PREFIX_FILE+" 2>/dev/null")
    os.system("rm -f " + TEMP_OUTPUT_FILE+" 2>/dev/null")

def signal_handler(signal, frame):
    print "You pressed Ctrl+C!"
    cleanup()
    exit(1)

def checkpatch():
    logging.info("Running checkpatch on workspace")
    ret=exprom_process.SystemExec("cpatch_raw.sh")
    logging.info("returncode=%d output:\n%s"%(ret['returncode'],ret['output']))
    return ret['returncode']

if __name__ == "__main__":
    # Make sure we start from clean environment
    cleanup()
    signal.signal(signal.SIGINT, signal_handler)
    optParser = OptionParser();
    configureParser(optParser)
    (options, args) = optParser.parse_args()

    if (len(args) > 0):
        optParser.print_help()
        exit(1)

    if options.verbose:
        VERBOSE_MODE=""

    if options.rom_suffix:
        ROM_SUFFIX=".rom"
    else:
        ROM_SUFFIX=".mrom"



    if not options.no_checkpatch:
        try:
            if checkpatch() !=0:
                logging.error("Please FIX checkpatch errors. Bailing out")
                exit(1)
        except Exception, e:
            '''
            Advanced option - ignoring error
            '''
            pass

    if options.countlines:
        print "/.autodirect/mtrswgwork/accounter_lines_of_code/bin/loc_count.py -P ipxe -T pxe -V 0.1 -p " + os.getcwd()
        rc = os.system ( "/.autodirect/mtrswgwork/accounter_lines_of_code/bin/loc_count.py -P ipxe -T pxe -V 0.1 -p " + os.getcwd())
        if rc != 0:
            print "Counter lines analyser error"
            exit(rc)
        exit(0)

    rc = checkOptions(options)
    if rc != 0:
        optParser.print_help()
        exit(1)

    if options.git:
        tmpDir = "./cloned_" + time.strftime("%d_%m_%Y") + "/"
        if (os.path.isdir(tmpDir)):
            print "Removing " + tmpDir
            os.system("rm -rf " + tmpDir)
        cloneRes = os.system("git clone " + options.git + " " + tmpDir)
        if (cloneRes != 0):
            print "Failed to clone git repository. Exiting... (Valid git? Any free space in " + os.getcwd() + ")"
            cleanup()
            exit(cloneRes)
        if (options.tag):
            cloneRes = os.system("cd " + tmpDir + " && pwd && git checkout " + options.tag + " && cd -")
            if cloneRes != 0:
                print "Failed to checkout to the requested tag. Removing Cloned directory " + tmpDir
                os.system("rm -rf " + tmpDir)
                cleanup()
                exit(cloneRes)
        os.chdir(tmpDir + "src/" )
        os.system("pwd")
    elif options.tag:
        print "Building from current directory. Ignoring tag..."
    
    ##################################################################
    ## From here we work on the current directory (local or cloned) ##
    ##################################################################

    if options.coverity:
        # Set coverity configure
        print "************     Coverity configuration Start    ************"
        rc = os.system (COVERITY_CONFIGURE)
        if rc != 0:
            print "Coverity configuration error! Exiting..."
            cleanup()
            exit(1)
        print "************     Coverity configuration End    ************\n"


    # Create a backup files
    os.system("\cp -f " + GENERAL_FILE + " " + GENERAL_FILE + "_")
    os.system("\cp -f " + ROM_PREFIX_FILE + " " + ROM_PREFIX_FILE + "_")


    # Set general.h with version
    newlines = []
    oldlines = openFile(GENERAL_FILE)
    for line in oldlines:
        newlines.append(line.replace('__BLD_VERSION__', '__BUILD_VERSION__ \"' + options.version + '\"'))
    writeOutput("".join(newlines), GENERAL_FILE)

    if (options.device == "ALL"):
        requestedDevices = DEVICES
    else:
        requestedDevices = {options.device: DEVICES[options.device]}

    # Create destination directory
    buildDir = "./build_" + options.version
    if (os.path.isdir(buildDir)):
        print "Removing " + buildDir
        rc = os.system("rm -rf " + buildDir)
        if rc != 0:
            print "Failed to remove directory " + buildDir + ". Exiting..."
            cleanup()
            exit(rc);
    print "Creating output directory: " + buildDir
    rc = os.system("mkdir " + buildDir);
    if rc != 0:
        print "Failed to create directory " + buildDir + ". Exiting..."
        cleanup()
        exit(rc);

    # Build versions
    outputDir = buildDir
    outputDir += "/"
    if (not os.path.isdir(outputDir)):
        rc = os.system("mkdir -p " + outputDir);
        if rc != 0:
            print "Failed to create directory " + outputDir + ". Exiting..."
            cleanup()
            exit(rc);
    rc = build(options.version, requestedDevices, outputDir, options.coverity, options.debug, options.test)
    if rc != 0:
        cleanup()
        exit(rc);

    cleanup()

    print "Done."
    exit(0)
